from typing import Optional
from pydantic.fields import Field
from pydantic.main import ModelMetaclass, BaseModel


class ObjectMetaClass(ModelMetaclass):

     def __new__(mcls, name, bases, attrs):
        
        slots = set(attrs.pop('__slots__', tuple())) # The slots from: https://github.com/samuelcolvin/pydantic/issues/655#issuecomment-610900376
        for base in bases:
            if hasattr(base, '__slots__'):
                slots.update(base.__slots__)

        if '__dict__' in slots:
            slots.remove('__dict__')
        attrs['__slots__'] = tuple(slots)
        return super(ObjectMetaClass, mcls).__new__(mcls, name, bases, attrs)



class ArkitektObject(BaseModel, metaclass=ObjectMetaClass):
    TYPENAME: str = Field(None, alias='__typename')

    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(self.__dict__, indent=4, width=1)

    def _repr_html_(self):
        
        def buildTable(attributes):
            tablestring = "<table>"
            for key, value in attributes.items():
                tablestring = tablestring + (f"""
                    <tr>
                        <td>{key.capitalize()}</td>
                        <td>{value}</td>
                    </tr>
                """)
            return tablestring + "</table>"



        return f"""
            <p> Instance of {self.__class__.__name__} <p>
            {buildTable(self.__dict__)}
        """

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)


from bergen.actors.base import Actor
from bergen.handlers import *
from bergen.utils import *
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor


import asyncio

class ClassicActor(Actor):
    pass


class ClassicFuncActor(ClassicActor):
    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self, assign_handler: AssignHandler, args, kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        result = await self.assign(assign_handler, args, kwargs)
        try:
            shrinked_returns = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
            await assign_handler.pass_return(shrinked_returns)
        except Exception as e:
            await assign_handler.pass_exception(e)


class ClassicGenActor(ClassicActor):

    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self, assign_handler: AssignHandler, args, kwargs):
        raise NotImplementedError("This needs to be overwritten in order to work")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        async for result in self.assign(assign_handler, args, kwargs):
            lastresult = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
            await assign_handler.pass_yield(lastresult)

        await assign_handler.pass_done()






from bergen.schema import DataModel, DataPoint, WardSettings
from bergen.wards.default import GraphQLWard
from bergen.auths.base import BaseAuthBackend
from typing import Callable, Dict
from bergen.enums import DataPointType
import logging
import asyncio
from bergen.console import console
logger = logging.getLogger(__name__)

ward_registry = None

class WardRegistry(object):


    def __init__(self) -> None:
        self.distinctWardMap = {}
        self.distinctNegotiationMap: Dict[str, dict] = {}


        # For the Identifier on the Models
        self.identifierWardMap= {}
        self.builders =  {
                # Default Builders for standard
                DataPointType.GRAPHQL: GraphQLWard
        }

    @property
    def wards(self):
        return [ward for distinct, ward in self.distinctWardMap.items()]

    def set_base(self, ward):
        self.identifierWardMap["node"] = ward
        self.identifierWardMap["pod"] = ward
        self.identifierWardMap["template"] = ward


    def get_ward(self, identifier):
        assert identifier in self.identifierWardMap, f"No Model registered on Identifier {identifier}"
        return self.identifierWardMap[identifier]


    def register_model(self, model: DataModel):
        assert model.point.distinct in self.distinctWardMap, "We cannot register this Model because there exists no ward for it"
        self.identifierWardMap[model.identifier] = self.distinctWardMap[model.point.distinct]

    def create_ward(self, bergen, ward_setting: WardSettings):
        logger.info(f"Creating new Ward for Datapoint {ward_setting}")

        if ward_setting.type in self.builders:
            builder = self.builders[ward_setting.type]
            self.distinctWardMap[ward_setting.distinct]  = builder(bergen, ward_setting)
        else:
            raise NotImplementedError(f"We have no idea how to build the Ward of Type {ward_setting.type}")

    async def configure_wards(self):
        wards = [ward for distinct, ward in self.distinctWardMap.items()]
        distincts = [distinct for distinct, ward in self.distinctWardMap.items()]
        negotiation_results = await asyncio.gather(*[ward.configure() for ward in wards], return_exceptions=True)
        return {distinct: result for distinct, result in zip(distincts, negotiation_results)}



def get_ward_registry() -> WardRegistry:
    global ward_registry
    if ward_registry is None:
        ward_registry = WardRegistry()
    return ward_registry


from contextvars import ContextVar
from rich.console import ConsoleRenderable
from rich.panel import Panel
from rich.table import Table
from rich.live import Live
from bergen.console import console

current_monitor: ContextVar["Monitor"] = ContextVar('current_monitor', default=None)


class Monitor:

    def __init__(self, title="Monitor", log=False) -> None:
        """Monitor allows you to monitor the progress of what is happenening inside your application


        Args:
            title (str, optional): The Title of this Monitor (Is the Panels Title). Defaults to "Monitor".
            progress (bool, optional): Do you want to monitor the progress of assignments and reservations? Defaults to False.
        """
        self.columns = Table.grid(expand=True)
        self.columns.add_column()
        self.log = log
        self.panel = Panel(self.columns, title=title)
        self.live = Live(self.panel, refresh_per_second=4, console=console)


    def addRow(self, renderable: ConsoleRenderable):
        self.columns.add_row(renderable)

    def __aenter__(self):
        '''Convenience Method'''
        return self.__aenter__()

    def __aexit__(self,*args,**kwargs):
        self.__exit__(*args,**kwargs)

    def __enter__(self):
        current_monitor.set(self)
        self.live.__enter__()
        return self


    def __exit__(self, *args, **kwargs):
        self.live.__exit__(*args, **kwargs)
        current_monitor.set(None)





from bergen.query import TypedGQL


class ModelConfigurationError(Exception):
    pass


class BaseManager:

    def _call_meta(self, attribute, ward=None, **kwargs):
        method =  getattr(self.meta, attribute, None)
        assert method is not None, f"Please provide the {attribute} parameter in your ArnheimModel meta class for SchemaClass {self.model.__name__} "
        typed_gql: TypedGQL = method(self.model)    
        return typed_gql.run(ward=ward, variables=kwargs)



from pydantic.main import BaseModel
from ....messages.types import  ASSIGN, PROVIDE, UNASSIGN
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    with_progress: bool = False

class MetaModel(MessageMetaModel):
    '''The reference of the metamodel representats the assignation on the platform '''
    type: str = UNASSIGN
    extensions: Optional[MetaExtensionsModel]

class DataModel(MessageDataModel):
    assignation: str


class UnassignMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from bergen.actors.base import Actor
from bergen.handlers import *
from bergen.utils import *
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from bergen.actors.utils import *
import janus
import asyncio
import threading


class FunctionalActor(Actor):
    pass


class FunctionalFuncActor(FunctionalActor):
    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self, *args, **kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        bounce_context.set(assign_handler.message.meta.context)
        assign_handler_context.set(assign_handler)
        provide_handler_context.set(self.provide_handler)
        #
        result = await self.assign(*args, **kwargs)

        try:
            shrinked_returns = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
            await assign_handler.pass_return(shrinked_returns)
        except Exception as e:
            await assign_handler.pass_exception(e)


class FunctionalGenActor(FunctionalActor):

    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def assign(self,*args, **kwargs):
        raise NotImplementedError("This needs to be overwritten in order to work")

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        bounce_context.set(assign_handler.message.meta.context)
        assign_handler_context.set(assign_handler)
        provide_handler_context.set(self.provide_handler)

        try:
            async for result in self.assign(*args, **kwargs):
                lastresult = await shrinkOutputs(self.template.node, result) if self.shrinkOutputs else result
                await assign_handler.pass_yield(lastresult)

            await assign_handler.pass_done()
        except Exception as e:
            await assign_handler.pass_exception(e)


class FunctionalThreadedFuncActor(FunctionalActor):
    nworkers = 5

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def assign(self, *args, **kwargs):
        raise NotImplementedError("")


    async def iterate_queue(self, async_q, assign_handler: AssignHandler, provide_handler):
        while True:
            val = await async_q.get()
            action = val[0]
            if action == "log":
                await assign_handler.pass_log(val[1][0], level=val[1][1])
                async_q.task_done()
            if action == "return":
                await assign_handler.pass_return(val[1])
                async_q.task_done()
                break
            if action == "exception":
                await assign_handler.pass_exception(val[1])
                async_q.task_done()
                break  



    def _assign_threaded(self, args, kwargs, queue, context):
        queue_context.set(queue)
        bounce_context.set(context)
        try:
            result = self.assign(*args, **kwargs)
            lastresult = shrinkOutputsSync(self.template.node, result) if self.shrinkOutputs else result
            queue.put(("return", lastresult))
            queue.join()

        except Exception as e:
            console.print_exception()
            queue.put(("exception", e))
            queue.join()

        queue_context.set(None)
        bounce_context.set(None)


    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        queue = janus.Queue()

        try:
            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded, args, kwargs, queue.sync_q, assign_handler.message.meta.context)
            queuefut =  self.iterate_queue(queue.async_q, assign_handler, self.provide_handler)

            try:
                await asyncio.gather(threadedfut, queuefut)
                queue.close()
                await queue.wait_closed()
            except asyncio.CancelledError as e:
                console.log("Received Cancellation for task")

                if not threadedfut.done():
                    print("Sending request to Queue To Cancell")
                    threadedfut.cancel()

                try:
                    await threadedfut
                except asyncio.CancelledError as e:
                    print("Sucessfully Cancelled Thread")
                    raise e

        except Exception as e:
            await assign_handler.pass_exception(e)



class FunctionalThreadedGenActor(FunctionalActor):
    nworkers = 5

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def assign(self, *args, **kwargs):
        raise NotImplementedError("")  


    def _assign_threaded(self, args, kwargs, queue, event, context):
        queue_context.set(queue)
        bounce_context.set(context)
        try:
            if event.is_set(): 
                queue.put(("cancelled","Happy doneness")) 
                return

            for result in self.assign(*args, **kwargs):
                if event.is_set():
                    queue.put(("cancelled","Happy doneness"))
                    return
                    
                lastresult = shrinkOutputsSync(self.template.node, result) if self.shrinkOutputs else result
                queue.put(("yield", lastresult))
                queue.join()
        except Exception as e:
            console.print_exception()
            queue.put(("exception", e))

        queue.put(("done","Happy doneness"))
        queue_context.set(None)
        bounce_context.set(None)


    async def iterate_queue(self, async_q, assign_handler: AssignHandler, provide_handler):

        while True:
            val = await async_q.get()
            action = val[0]
            if action == "log":
                await assign_handler.pass_log(val[1][0], level=val[1][1])
                async_q.task_done()
            if action == "yield":
                await assign_handler.pass_yield(val[1])
                async_q.task_done()
            if action == "exception":
                await assign_handler.pass_exception(val[1])
                async_q.task_done()
                break
            if action == "done":
                await assign_handler.pass_done()
                async_q.task_done()
                break

        

    async def _assign(self, assign_handler: AssignHandler, args, kwargs):
        queue = janus.Queue()
        print(assign_handler.message.meta.context)

        event = threading.Event()

        try:
            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded, args, kwargs, queue.sync_q, event, assign_handler.message.meta.context)
            queuefut =  self.iterate_queue(queue.async_q, assign_handler, self.provide_handler)

            try:
                await asyncio.gather(threadedfut, queuefut)
                queue.close()
                await queue.wait_closed()
            except asyncio.CancelledError as e:
                console.log("Received Cancellation for task")

                if not threadedfut.done():
                    print("Sending request to Queue To Cancell")
                    event.set()
                    threadedfut.cancel()

                try:
                    await threadedfut
                except asyncio.CancelledError as e:
                    print("Sucessfully Cancelled Thread")
                    raise e

        except Exception as e:
            await assign_handler.pass_exception(e)


from asyncio.futures import Future
from bergen.messages.postman.log import LogLevel
from contextvars import Context
from bergen.messages.postman.reserve.reserve_transition import ReserveState
from bergen import messages
from bergen.contracts.exceptions import AssignmentException
from bergen.registries.client import get_current_client
from bergen.schema import Node, NodeType
from bergen.monitor import Monitor, current_monitor
from bergen.messages.postman.reserve.params import ReserveParams
from bergen.messages import *
from bergen.utils import *
from rich.table import Table
from rich.panel import Panel
import asyncio
import logging
from bergen.console import console

logger = logging.getLogger(__name__)

class UnknownMessageError(Exception):
    pass


class ReservationError(Exception):
    pass


class CouldNotReserveError(ReservationError):
    pass


class IncorrectStateForAssignation(ReservationError):
    pass




class Reservation:

    def __init__(self, node: Node,
        reference: str = None,
        provision: str = None, 
        monitor: Monitor = None,
        ignore_node_exceptions=False,
        transition_hook=None,
        with_log=False,
        enter_on=[ReserveState.ACTIVE], 
        exit_on=[ReserveState.ERROR, ReserveState.CANCELLED],
        context: Context =None,
        loop=None,
         **params) -> None:


        # Reference
        self.monitor = monitor or current_monitor.get()
        self.client = get_current_client()
        self.loop = loop or self.client.loop
        self._postman = self.client.postman


        # Reservation Params
        self.reference = reference or str(uuid.uuid4())
        self.provision = provision
        self.node = node
        self.params = ReserveParams(**params)
        self.with_log = with_log or (self.monitor.log if self.monitor else None)
        self.context = context # with_bounced allows us forward bounced checks


        if self.context:
            assert "can_forward_bounce" in self.client.auth.scopes, "In order to use with_bounced forwarding you need to have the can_forward_bounced scope"

        # Exception Mangement
        self.ignore_node_exceptions = ignore_node_exceptions
        self.critical_error = None

        # State management
        self.transition_hook = transition_hook
        self.exit_states = exit_on
        self.enter_states = enter_on
        self.current_state = None


        if self.monitor:
            self.monitor.addRow(self.build_panel())
            self._log = lambda message, level: self.table.add_row(level, message)
        else:
            self._log = lambda message, level, : console.log(f"Reservation {self.reference} {level}: {message}")

        pass
    
    def log(self, message: str, level: LogLevel=LogLevel.DEBUG):
        """Logs a Message

        The Logged Message will be display on the Monitor if running inside a Monitor
        and send to the logging output.

        Args:
            message (str): The Message
            level (LogLevel, optional): The LogLevel. Defaults to LogLevel.DEBUG.
        """
        self._log(message, level)



    def build_panel(self) -> Panel:
        """Builds and returns a panel for a Monitor of this Reservation

        Returns:
            Panel: The Build Panel
        """
        heading_information = Table.grid(expand=True)
        heading_information.add_column()
        heading_information.add_column(style="green")

        reserving_table = Table(title=f"[bold green]Reserving on ...", show_header=False)

        for key, value in self.params.dict().items():
            reserving_table.add_row(key, str(value))

        heading_information.add_row(self.node.__rich__(), reserving_table)

        self.table = Table()
        self.table.add_column("Level")
        self.table.add_column("Message")

        columns = Table.grid(expand=True)
        columns.add_column()

        columns.add_row(heading_information)
        columns.add_row(self.table)

        return Panel(columns, title="Reservation")


    async def assign_async(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, with_log=True, context=None, **kwargs):
        assert self.node.type == NodeType.FUNCTION, "You cannot assign to a Generator Node, use the stream Method!"

        if self.current_state in self.exit_states:
            raise IncorrectStateForAssignation(f"Current State {self.current_state} is an Element of Exit States {self.exit_states}")

        self.log(f"Streaming!", level=LogLevel.INFO)
        message_queue = asyncio.Queue()

        shrinked_args, shrinked_kwargs = await shrinkInputs(self.node, args, kwargs) if not bypass_shrink else (args, kwargs)
        assign_reference = await self._postman.stream_assign_to_queue(message_queue, self.reference, shrinked_args=shrinked_args, shrinked_kwargs=shrinked_kwargs, with_log=with_log, context=context or self.context)

        try:
            while True:
                message = await message_queue.get()

                if isinstance(message, AssignReturnMessage):
                    outs = await expandOutputs(self.node, message.data.returns) if not bypass_expand else message.data.returns    
                    return outs

                elif isinstance(message, AssignCancelledMessage):
                    raise AssignmentException(f"Assignment was cancelled from a different Agent: ID: {message.data.canceller}")

                elif isinstance(message, AssignLogMessage):
                    self.log(message.data.message, message.data.level)

                elif isinstance(message, AssignCriticalMessage):
                    raise AssignmentException(message.data.message)

                elif isinstance(message, AssignYieldsMessage):
                    raise AssignmentException("Received a Yield from a Node that should never yield! CRITICAL PROTOCOL EXCEPTION")

                else:
                    raise UnknownMessageError(message)


        except asyncio.CancelledError as e:
            self.log("Assigment Required Cancellation", level=LogLevel.INFO)

            unassign_reference = await self._postman.send_unassign(assign_reference, context=context)

            while True:
                message = await message_queue.get()
                if isinstance(message, AssignCancelledMessage):

                    if message.data.canceller != unassign_reference:
                        self.log("Canceller does not match our Cancellation Request, Race Condition?")
                    raise e

                else:
                    print("Raced Condition",message)



    async def stream(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, context=None, with_log=True, **kwargs):
        assert self.node.type == NodeType.GENERATOR, "You cannot stream a Function Node, use the assign Method!"

        if self.current_state in self.exit_states:
            raise IncorrectStateForAssignation(f"Current State {self.current_state} is an Element of Exit States {self.exit_states}")

        self.log(f"Assigning!", level=LogLevel.INFO)
        message_queue = asyncio.Queue()

        shrinked_args, shrinked_kwargs = await shrinkInputs(self.node, args, kwargs) if not bypass_shrink else (args, kwargs)
        assign_reference = await self._postman.stream_assign_to_queue(message_queue, self.reference, shrinked_args=shrinked_args, shrinked_kwargs=shrinked_kwargs, with_log=with_log, context=context or self.context)

        try:
            while True:
                message = await message_queue.get()

                if isinstance(message, AssignYieldsMessage):
                    outs = await expandOutputs(self.node, message.data.returns) if not bypass_expand else message.data.returns    
                    yield outs

                elif isinstance(message, AssignDoneMessage):   
                    break

                elif isinstance(message, AssignCancelledMessage):
                    raise AssignmentException(f"Assignment was cancelled from a different Agent: ID: {message.data.canceller}")

                elif isinstance(message, AssignLogMessage):
                    self.log(message.data.message, message.data.level)

                elif isinstance(message, AssignCriticalMessage):
                    raise AssignmentException(message.data.message)

                elif isinstance(message, AssignReturnMessage):
                    raise AssignmentException("Received a Return from a Node that should never return! CRITICAL PROTOCOL EXCEPTION")

                else:
                    raise UnknownMessageError(message)


        except asyncio.CancelledError as e:
            self.log("Assigment Required Cancellation", level=LogLevel.INFO)

            unassign_reference = await self._postman.send_unassign(assign_reference, context=context)

            while True:
                message = await message_queue.get()

                if isinstance(message, AssignCancelledMessage):

                    if message.data.canceller != unassign_reference:
                        self.log("Canceller does not match our Cancellation Request, Race Condition?")
                    raise e

                elif isinstance(message, AssignLogMessage):
                    self.log(message.data.message, message.data.level)

                else:
                    print("Raced Condition",message)



    async def stream_worker(self, queue: asyncio.Queue):
        try:
            self.reference = await self._postman.stream_reserve_to_queue(queue, node_id=self.node.id, provision=self.provision, params_dict=self.params.dict(), with_log=self.with_log, context=self.context)
            print(self.reference)
            while True: 
                message = await queue.get()

                if isinstance(message, ReserveCriticalMessage):
                    if self.enter_future.done():
                        self.log(f"We have transitioned to a critical State {message.data.message}")
                        self.current_state = ReserveState.CRITICAL
                    else:
                        self.enter_future.set_exception(Exception(message.data.message))



                elif isinstance(message, ReserveTransitionMessage):
                        # Once we acquire a reserved resource our contract (the inner part of the context can start)
                        self.current_state  = message.data.state
                        if self.transition_hook: await self.transition_hook(self, message.data.state)
                        
                        if self.current_state in self.exit_states:
                            if not self.is_closing: self.log(f"Received Exitstate: {message.data.state}. Closing reservation at next assignment", level=LogLevel.CRITICAL)
                            return # This means we do no longer need to here about stuff

                        if self.current_state in self.enter_states:
                            if self.enter_future.done():
                                self.log("We are already entered.")
                            else:
                                self.enter_future.set_result(message.meta.reference)

                else:
                    console.log(message)

                queue.task_done()
        
        except Exception as e: 
            console.print_exception()
            if not self.enter_future.done():
                self.enter_future.set_exception(e)
            raise e


    async def cancel(self):
        await self._postman.send_unreserve(self.reference, context=self.context)

    async def start(self):
        return await self.__aenter__()

    async def end(self):
        await self.__aexit__()

    async def __aenter__(self):
        self.reservation_queue = asyncio.Queue()
        self.is_closing = False
        
        self.enter_future = self.loop.create_future()
        self.stream_task = self.loop.create_task(self.stream_worker(self.reservation_queue))

        print("Reached Here")

        try:
            self.enter_state = await self.enter_future
            return self

        except Exception as e:
            raise CouldNotReserveError(f"Could not Reserve Reservation {self.reference} for Node {self.node}") from e


    async def __aexit__(self, type, value, traceback):
        self.is_closing = True
        await self.cancel()

        if not self.stream_task.done():
            #TODO: Maybe make a Connection this
            self.is_closing = True
            await self.stream_task
            self.log("Reservation sucessfully unreserved", level=LogLevel.INFO)
        
        if value is not None and isinstance(value, asyncio.CancelledError): 
            print("Raising cancellation")
            raise value # passing our cancellation error


        

    
    def assign(self, *args, bypass_shrink=False, bypass_expand=False, persist=True, **kwargs):
        return self.assign_async(*args, bypass_shrink=bypass_shrink, bypass_expand=bypass_expand, persist=persist, **kwargs)

    def __enter__(self):
        self.in_sync = True
        future = self.loop.run_until_complete(self.__aenter__())
        return future

    def __exit__(self, *args, **kwargs):
        return  self.loop.run_until_complete(self.__aexit__(*args, **kwargs))



    




       




from typing import Type
from bergen.types.node.widgets.querywidget import QueryWidget
from bergen.types.node.ports.returns.base import BaseReturnPort
from bergen.types.model import ArnheimModel

class ModelReturnPort(BaseReturnPort):

  def __init__(self, modelMeta,**kwargs) -> None:
      self.modelMeta = modelMeta
      super().__init__(**kwargs)

  def serialize(self):
      return {**super().serialize(),"identifier" : self.modelMeta.identifier}


from bergen.messages.postman.unprovide.unprovide_critical import UnprovideCriticalMessage
from bergen.messages.postman.unprovide.unprovide_log import UnprovideLogMessage
from bergen.messages.postman.log import LogLevel, LogDataModel
from bergen.messages.postman.unprovide.unprovide_done import UnprovideDoneMessage
from bergen.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from bergen.messages.postman.unreserve.unreserve_critical import UnreserveCriticalMessage
from bergen.messages.postman.reserve.reserve_critical import ReserveCriticalMessage
from bergen.schema import Template
from bergen.messages.postman.assign.assign_done import AssignDoneMessage
from bergen.debugging import DebugLevel
from bergen.handlers.base import ContractHandler
from bergen.messages import BouncedUnreserveMessage
from bergen.console import console


class UnprovideHandler(ContractHandler[BouncedUnprovideMessage]):

    @property
    def bounced(self):
        return self.message.meta.token

    @property
    def meta(self):
        return {"extensions": self.message.meta.extensions, "reference": self.message.meta.reference}

    async def pass_log(self, message, level=LogLevel.INFO):
        log_message = UnprovideLogMessage(data={"message": message, "level": level, "provision": self.message.data.provision}, meta=self.meta)
        return await self.connection.forward(log_message)   

    async def pass_done(self):
        done_message = UnprovideDoneMessage(data={"provision": self.message.data.provision}, meta=self.meta)
        return await self.connection.forward(done_message) #Contrary to a Done signal we are the provision just to the backend, because it will take care of it

    async def pass_error(self, exception):
        critical_message = UnprovideErrorMessage(data={"message": str(exception), "type": exception.__class__.__name__}, meta=self.meta)
        console.print_exception()
        return await self.connection.forward(critical_message)

    async def pass_critical(self, exception):
        critical_message = UnprovideCriticalMessage(data={"message": str(exception), "type": exception.__class__.__name__}, meta=self.meta)
        console.print_exception()
        return await self.connection.forward(critical_message)



from abc import ABC, abstractmethod
import asyncio
from bergen.managers.base import BaseManager
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Type
from pydantic.fields import Field
from pydantic.main import BaseModel, ModelMetaclass
from bergen.query import DelayedGQL, GQL, TypedGQL

class ArnheimModelException(Exception):
    pass

class ModelDoesNotExistError(ArnheimModelException):
    pass

class ArnheimModelConfigurationError(ArnheimModelException):
    pass


ModelType = TypeVar("ModelType", bound="ArnheimModel", covariant=True)

class ArnheimModelManager(BaseManager, Generic[ModelType]):

    def __init__(self, model: ModelType = None, meta = None) -> None:
        self.model = model
        self.meta = meta
        super().__init__()

    def _call_meta(self, attribute, **kwargs):
        from bergen.registries.client import get_current_client
        client = get_current_client()
        from bergen.types.utils import parse_kwargs
        method =  getattr(self.meta, attribute, None)
        assert method is not None, f"Please provide the {attribute} parameter in your ArnheimModel meta class "
        typed_gql: TypedGQL = method(self.model)
        if client.loop.is_running():
            return asyncio.run_coroutine_threadsafe(typed_gql.run(variables=parse_kwargs(kwargs)), client.loop).result()
        else:
            return client.loop.run_until_complete(typed_gql.run(variables=parse_kwargs(kwargs)))

    def __getattr__(self, name: str) -> ModelType:
        def function(**kwargs):
            return self._call_meta(name, **kwargs)
        return function

    def get(self, ward=None, **kwargs) -> ModelType:
        """Gets an instance of this Model from the Dataprovider registered with Arnheim,

        Args:
            ward ([type], optional): A custom ward to be used. If you want to fetch with a different Dataprovider than the one Bergen provides you. Defaults to None.

        Returns:
            ModelType: The returned Instance
        """
    
        return self._call_meta("get", ward=ward, **kwargs)
        
    def create(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("create", ward=ward, **kwargs)

    def filter(self, ward=None, **kwargs) -> List[ModelType]:
        return self._call_meta("filter", ward=ward, **kwargs)

    def update(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("update", ward=ward, **kwargs)

    def all(self, ward=None):
        return self._call_meta("filter", ward=ward)

    def __call__(self, model=None, meta=None):
        self.model = model
        self.meta = meta
        return self


class ArnheimAsyncModelManager(BaseManager, Generic[ModelType]):

    def __init__(self, model: ModelType = None, meta = None) -> None:
        self.model = model
        self.meta = meta
        super().__init__() 

    async def _call_meta(self, attribute, ward=None, **kwargs):
        from bergen.types.utils import parse_kwargs
        method =  getattr(self.meta, attribute, None)
        assert method is not None, f"Please provide the {attribute} parameter in your ArnheimModel meta class "
        typed_gql: TypedGQL = method(self.model)    
        return await typed_gql.run(ward=ward, variables=parse_kwargs(kwargs))


    def __getattr__(self, name: str) -> ModelType:
        async def function(**kwargs):
            return await self._call_meta(name, ward=self.get_ward(), **kwargs)
        return function

    async def get(self, ward=None, **kwargs) -> ModelType:
        return await self._call_meta("get", ward=ward, **kwargs)
        
    async def create(self, ward=None, **kwargs) -> ModelType:
        return await self._call_meta("create", ward=ward, **kwargs)

    async def filter(self, ward=None, **kwargs) -> List[ModelType]:
        return await self._call_meta("filter", ward=ward, **kwargs)


    async def fetch(self, gql, ward=None, **kwargs) -> List[ModelType]:
        from bergen.types.utils import parse_kwargs
        typed_gql = DelayedGQL(gql)(self.model)
        return await typed_gql.run(ward=ward, variables=parse_kwargs(kwargs))


    async def update(self, ward=None, **kwargs) -> ModelType:
        return await self._call_meta("update", ward=ward, **kwargs)

    async def all(self, ward=None):
        return await self._call_meta("filter", ward=ward)

    def __call__(self, model=None, meta=None):
        self.model = model
        self.meta = meta
        return self


class ArnheimModelMeta(ModelMetaclass):

    def __new__(mcls, name, bases, attrs):
        
        slots = set(attrs.pop('__slots__', tuple())) # The slots from: https://github.com/samuelcolvin/pydantic/issues/655#issuecomment-610900376
        for base in bases:
            if hasattr(base, '__slots__'):
                slots.update(base.__slots__)

        if '__dict__' in slots:
            slots.remove('__dict__')
        attrs['__slots__'] = tuple(slots)

        mcls.overriden_manager = attrs.pop("objects") if "objects" in attrs else None
        mcls.overriden_async_manager = attrs.pop("asyncs") if "asyncs" in attrs else None
        return super(ArnheimModelMeta, mcls).__new__(mcls, name, bases, attrs)


    @property
    def objects(cls: Type[ModelType]) -> ArnheimModelManager[ModelType]:
        return cls.__objects

    @property
    def asyncs(cls: Type[ModelType]) -> ArnheimAsyncModelManager[ModelType]:
        return cls.__asyncs


    def __init__(self, name, bases, attrs):
        super(ArnheimModelMeta, self).__init__(name, bases, attrs)
        if attrs["__qualname__"] != "ArnheimModel":
            # This gets allso called for our Baseclass which is abstract
            meta = attrs["Meta"] if "Meta" in attrs else None
            assert meta is not None, "Please provide a Meta class in your Arnheim Model"
            
            managerClass = self.overriden_manager
            if managerClass:
                self.__objects = managerClass(model=self,meta=meta)
            else:
                self.__objects = ArnheimModelManager[ModelType](model=self,meta=meta)

            asyncmanagerClass = self.overriden_async_manager 
            if asyncmanagerClass:
                self.__asyncs = asyncmanagerClass(model=self,meta=meta)
            else:
                self.__asyncs = ArnheimAsyncModelManager[ModelType](model=self,meta=meta)


            
            overwriteType = getattr(meta, "overwrite_default", False)
            register = getattr(meta, "register", True)
            if register:
                from bergen.registries.matcher import get_current_matcher
                identifier = getattr(meta, "identifier", None)
                assert identifier is not None, f"Please provide identifier in your Meta class to register the Model {attrs['__qualname__']}, or specifiy register==False"
                get_current_matcher().registerModelForIdentifier(identifier, self, overwrite=overwriteType)



class ArnheimModel(BaseModel, metaclass=ArnheimModelMeta):
    TYPENAME: str = Field(None, alias='__typename')
    id: Optional[int]

    @classmethod
    def get_ward(cls):
        try:
            identifier = cls.Meta.identifier
        except Exception as e:
            raise ArnheimModelConfigurationError(f"Make soure your Model {cls.__name__}overwrites Meta identifier: {e}")

        from bergen.registries.ward import get_ward_registry
        return get_ward_registry().get_ward(identifier=identifier)

    @classmethod
    def getMeta(cls):
        return cls.Meta

    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(self.__dict__, indent=4, width=1)

    def _repr_html_(self):
        
        def buildTable(attributes: Dict):
            tablestring = "<table>"
            for key, value in attributes.items():
                tablestring = tablestring + (f"""
                    <tr>
                        <td>{key.capitalize()}</td>
                        <td>{value}</td>
                    </tr>
                """)
            return tablestring + "</table>"



        return f"""
            <p> Instance of {self.__class__.__name__} <p>
            {buildTable(self.__dict__)}
        """

    def _repr_list_(self):
        string = f"{self.__class__.__name__}("
        if self.id is not None: string += str(self.id) + ", "
        if self.name is not None: string += str(self.name) +", "
        string = string[:-2] + ")"
        return string


    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)



from abc import abstractmethod
from abc import ABC
import asyncio
from bergen.auths.base import BaseAuthBackend
from bergen.schema import DataPoint, WardSettings
from bergen.query import  TypedGQL
from typing import TypeVar
from bergen.console import console


T = TypeVar("T")

class WardException(Exception):
    pass

class GraphQLException(WardException):
    pass

class ConnectionError(Exception):
    pass

class TokenExpired(ConnectionError):
    pass


class BaseWard(ABC):

    def __init__(self, client, loop=None):
        self.loop = loop or client.loop or asyncio.get_event_loop()
        self.client = client
        self.auth: BaseAuthBackend = client.auth
        assert self.auth.access_token is not None, "Cannot create a Ward without acquiring a Token first"
        self._headers = {"Authorization": f"Bearer {self.auth.access_token}"}

    @abstractmethod
    async def connect(self):
        """Everytime we need to reastablish a connection because of a Token Refersh"""
        pass

    @abstractmethod
    async def configure(self):
        """The initial connection attempt (also negotiation can happen here)"""
        pass


    @abstractmethod
    async def disconnect(self):
        """CleanUp"""
        pass


    async def run(self, gql: TypedGQL, variables: dict = {}):
        try:
            return await self.pass_async(gql, variables=variables)

        except TokenExpired:
            console.print_exception()

            self.auth.refetch()
            self._headers = {"Authorization": f"Bearer {self.auth.access_token}"}

            await self.disconnect()
            await self.connect()

            return await self.pass_async(gql, variables=variables)

        except:
            console.print_exception(show_locals=True)
            raise



    @abstractmethod
    def pass_async(self, gql: TypedGQL, variables: dict = {}):
        return gql.cls(**{})




class ServiceWard(BaseWard):

    def __init__(self, client, settings: WardSettings, loop=None):
        self.distinct = settings.distinct
        self.needs_negotiation = settings.needsNegotiation
        self.host = settings.host or client.config.host
        self.port = settings.port or client.config.port
        self.protocol = "https" if settings.secure or client.config.secure else "http"
        super().__init__(client, loop=loop)

    async def configure(self):
        try:
            await self.connect()
            if self.needs_negotiation: 
                return await self.negotiate()
            return 
        except:
            console.print_exception()
            raise ConnectionError(f"Ward {self.distinct}: Connection to {self.host}:{self.port} on {self.port} Failed")



class MainWard(BaseWard):

    def __init__(self, client, loop=None):
        self.host = client.config.host
        self.port = client.config.port
        self.protocol = "https" if client.config.secure else "http"
        super().__init__(client, loop=loop)

    async def configure(self):
        try:
            await self.connect()
        except:
            console.print_exception()
            raise ConnectionError(f"BaseWard: Connection to {self.host}:{self.port} on {self.port} Failed")



import asyncio
from bergen.legacy.utils import get_running_loop
from typing import Any
from bergen.contracts.reservation import Reservation
from bergen.contracts.interaction import Interaction

from bergen.monitor.monitor import Monitor
from bergen.messages.postman.reserve.bounced_reserve import ReserveParams
from bergen.schema import AssignationParams, Node, NodeType
from bergen.registries.client import get_current_client
from bergen.contracts import Reservation
from aiostream import stream
from tqdm import tqdm
import textwrap
import logging
from rich.table import Table
from rich.table import Table


logger = logging.getLogger(__name__)

class AssignationUIMixin:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ui = None


    def askInputs(self, **kwargs) -> dict:
        widget = self.getWidget(**kwargs) # We have established a ui
        if widget.exec_():
            return widget.parameters
        else:
            return None


    def getWidget(self, **kwargs):
        try:
            from bergen.ui.assignation import AssignationUI
            if not self._ui:
                self._ui = AssignationUI(self.inputs, **kwargs)
            return self._ui
        except ImportError as e:
            raise NotImplementedError("Please install PyQt5 in order to use interactive Widget based parameter query")



class NodeExtender(AssignationUIMixin):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        client = get_current_client()

        self._postman = client.postman
        self._loop = client.loop


    def interactive(self) -> Interaction:
        return Interaction(self)

    def reserve(self, loop=None, monitor: Monitor = None, ignore_node_exceptions=False, bounced=None, **params) -> Reservation:
        return Reservation(self, loop=loop, monitor=monitor, ignore_node_exceptions=ignore_node_exceptions, bounced=bounced, **params)


    def _repr_html_(self: Node):
        string = f"{self.name}</br>"

        for arg in self.args:
            string += "Args </br>"
            string += f"Port: {arg._repr_html_()} </br>"

        for kwarg in self.kwargs:
            string += "Kwargs </br>"
            string += f"Port: {kwarg._repr_html_()} </br>"


        return string


    async def assign(self, *args, reserve_params: dict = {}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            return await res.assign(*args, **kwargs)

    async def stream(self, *args, reserve_params: dict = {}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            async for result in res.stream(*args, **kwargs):
                yield result

        


    def __call__(self, *args: Any,  reserve_params: ReserveParams = None, **kwargs) -> Any:
        try:
            loop = get_running_loop()
        except RuntimeError:
            assert self.type == NodeType.FUNCTION, "Can only call node functions syncronoicouly"
            self._loop.run_until_complete()
        else:
            if self.type == NodeType.GENERATOR:
                return self.stream(*args, **kwargs)
            if self.type == NodeType.FUNCTION:
                return self.assign(*args, **kwargs)


    def __rich__(self):
        my_table = Table(title=f"Node: {self.name}", show_header=False)

        my_table.add_row("ID", str(self.id))
        my_table.add_row("Package", self.package)
        my_table.add_row("Interface", self.interface)
        my_table.add_row("Type", self.type)

        return my_table


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bergen',
 'bergen.actors',
 'bergen.auths',
 'bergen.auths.backend',
 'bergen.auths.code',
 'bergen.auths.code.widgets',
 'bergen.auths.implicit',
 'bergen.auths.implicit.widgets',
 'bergen.auths.legacy',
 'bergen.auths.messages',
 'bergen.auths.messages.postman',
 'bergen.auths.messages.postman.assign',
 'bergen.auths.messages.postman.provide',
 'bergen.auths.messages.postman.reserve',
 'bergen.auths.messages.postman.unassign',
 'bergen.auths.messages.postman.unprovide',
 'bergen.auths.messages.postman.unreserve',
 'bergen.clients',
 'bergen.config',
 'bergen.contracts',
 'bergen.entertainer',
 'bergen.extenders',
 'bergen.graphql',
 'bergen.graphql.types',
 'bergen.handlers',
 'bergen.hookable',
 'bergen.legacy',
 'bergen.managers',
 'bergen.messages',
 'bergen.messages.postman',
 'bergen.messages.postman.assign',
 'bergen.messages.postman.provide',
 'bergen.messages.postman.reserve',
 'bergen.messages.postman.unassign',
 'bergen.messages.postman.unprovide',
 'bergen.messages.postman.unreserve',
 'bergen.monitor',
 'bergen.postmans',
 'bergen.provider',
 'bergen.queries',
 'bergen.queries.delayed',
 'bergen.registries',
 'bergen.schemas.arkitekt',
 'bergen.schemas.arkitekt.mutations',
 'bergen.schemas.herre',
 'bergen.types',
 'bergen.types.node',
 'bergen.types.node.ports',
 'bergen.types.node.ports.arg',
 'bergen.types.node.ports.kwarg',
 'bergen.types.node.ports.returns',
 'bergen.types.node.widgets',
 'bergen.ui',
 'bergen.ui.widgets',
 'bergen.wards',
 'bergen.wards.bare',
 'bergen.wards.gql']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'aiostream>=0.4.1,<0.5.0',
 'docstring-parser>=0.7.3,<0.8.0',
 'janus>=0.6.1,<0.7.0',
 'namegenerator>=1.0.6,<2.0.0',
 'nest-asyncio>=1.5.1,<2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=10.1.0,<11.0.0',
 'tqdm>=4.56.1,<5.0.0',
 'websockets>=8.1,<9.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.8"': ['contextvars>=2.2,<3.0'],
 'gql': ['gql[all]>=3.0.0a5,<4.0.0'],
 'pyqt': ['PyQt5>=5.15.4,<6.0.0', 'PyQtWebEngine>=5.15.4,<6.0.0'],
 'pyside': ['PyQtWebEngine>=5.15.4,<6.0.0', 'PySide2>=5.15.2,<6.0.0']}

setup_kwargs = {
    'name': 'bergen',
    'version': '0.4.75',
    'description': 'A python client for the Arnheim Framework',
    'long_description': '# Bergen\n\n### Idea\n\nBergen is the API-Client for the Arnheim Framework\n\n \n### Prerequisites\n\nBergen only works with a running Arnheim Instance (in your network or locally for debugging).\n\n### Usage\n\nIn order to initialize the Client you need to connect it as a Valid Application with your Arnheim Instance\n\n```python\nclient = Bergen(host="p-tnagerl-lab1",\n    port=8000,\n  client_id="APPLICATION_ID_FROM_ARNHEIM", \n  client_secret="APPLICATION_SECRET_FROM_ARNHEIM",\n  name="karl",\n)\n```\n\nIn your following code you can simple query your data according to the Schema of the Datapoint\n\n```python\nfrom bergen.schema import Node\n\nnode = Node.objects.get(id=1)\nprint(node.name)\n\n```\n\n## Access Data from different Datapoints\n\nThe Arnheim Framework is able to provide data from different Data Endpoints through a commong GraphQL Interface\n. This allows you to access data from various different storage formats like Elements and Omero and interact without\nknowledge of their underlying api.\n\nEach Datapoint provides a typesafe schema. Arnheim Elements provides you with an implemtation of that schema.\n\n## Provide a Template for a Node\n\nDocumentation neccesary\n\n\n### Testing and Documentation\n\nSo far Bergen does only provide limitedunit-tests and is in desperate need of documentation,\nplease beware that you are using an Alpha-Version\n\n\n### Build with\n\n- [Arnheim](https://github.com/jhnnsrs/arnheim)\n- [Pydantic](https://github.com/jhnnsrs/arnheim)\n\n\n#### Features\n\n- Scss\n- [Domain-style](https://github.com/reactjs/redux/blob/master/docs/faq/CodeStructure.md) for code structure\n- Bundle Size analysis\n- Code splitting with [react-loadable](https://github.com/jamiebuilds/react-loadable)\n\n\n## Roadmap\n\nThis is considered pre-Alpha so pretty much everything is still on the roadmap\n\n\n## Deployment\n\nContact the Developer before you plan to deploy this App, it is NOT ready for public release\n\n## Versioning\n\nThere is not yet a working versioning profile in place, consider non-stable for every release \n\n## Authors\n\n* **Johannes Roos ** - *Initial work* - [jhnnsrs](https://github.com/jhnnsrs)\n\nSee also the list of [contributors](https://github.com/your/project/contributors) who participated in this project.\n\n## License\n\nAttribution-NonCommercial 3.0 Unported (CC BY-NC 3.0) \n\n## Acknowledgments\n\n* EVERY single open-source project this library used (the list is too extensive so far)',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jhnnsrs/bergen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<3.10',
}


setup(**setup_kwargs)



from bergen.models import Node
import asyncio




def use(package=None, interface=None) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    loop = asyncio.get_event_loop()
    if loop.is_running():
        return Node.asyncs.get(package=package, interface=interface)
    else:
        return Node.objects.get(package=package, interface=interface)




from .postman.provide import *
from .postman.reserve import *
from .postman.unprovide import *
from .postman.unreserve import *
from .postman.assign import *
from .postman.unassign import *
from .exception import *
from .base import *


import asyncio
from abc import ABC, abstractmethod
from asyncio.futures import Future
from bergen.messages.postman.log import LogLevel
from bergen.handlers.unprovide import UnprovideHandler
from bergen.clients.base import BaseBergen
from bergen.handlers.base import Connector
from bergen.messages import *
from bergen.hookable.base import Hookable
import logging
from typing import Dict, Type
from bergen.models import Pod
from bergen.actors.base import Actor
from functools import partial
from bergen.console import console
from bergen.legacy.utils import create_task



logger = logging.getLogger(__name__)

class ProtocolError(Exception):
    pass

class TaskAlreadyDoneError(ProtocolError):
    pass




class BaseEntertainer(Hookable):
    ''' Is a mixin for Our Bergen '''
    def __init__(self, client: BaseBergen, raise_exceptions_local=False, loop=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.provisions = {}
        self.raise_exceptions_local = raise_exceptions_local
        self.loop = loop or client.loop or asyncio.get_event_loop()
        self.client = client

        self.connector = Connector(self)

        self.template_id_actorClass_map: Dict[str, Type[Actor]] = {}

        self.provision_actor_map: Dict[str, Actor] = {}
        self.provision_actor_run_map: Dict[str, asyncio.Task]= {}
        self.provision_actor_queue_map: Dict[str, asyncio.Queue] = {}


        self.unprovide_futures: Dict[str, Future] = {}
        self.provide_futures: Dict[str, Future] = {}

        self.all_pod_assignments = {}
        self.all_pod_reservations = {}

        self.entertainments = {}
        self.assignments = {}

        self.pending = []

        self.tasks = []

    @abstractmethod
    async def connect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def disconnect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def activateProvision(self):
         raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def deactivateProvision(self):
         raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def forward(self, message: MessageModel):
        raise NotImplementedError("Overwrite this in your Protocol")


    async def on_message(self, message: MessageModel):
        
        if isinstance(message, ProvideDoneMessage): 
            # As we send the Activate Provide request to the Platform we will get a Return Statement
            logger.info("Arkitekt acknowledged the creation of a Pod for our Actor [this has no consequences??]")
            self.provide_futures[message.meta.reference].set_result(message.data.pod)

        elif isinstance(message, UnprovideDoneMessage): 
            # As we send the Activate Provide request to the Platform we will get a Return Statement
            logger.info("Arkitekt acknowledged the deletion of the Pod our our former Actor")
            self.unprovide_futures[message.meta.reference].set_result(message.data)


        elif isinstance(message, BouncedForwardedAssignMessage):
            assert message.data.provision is not None, "Received assignation that had no Provision???"
            assert message.data.provision in self.provision_actor_queue_map, f"Provision not entertained {message.data.provision} {self.provision_actor_queue_map}"
            await self.provision_actor_queue_map[message.data.provision].put(message)
            self.all_pod_assignments[message.meta.reference] = message.data.provision # Run in parallel

        elif isinstance(message, BouncedUnassignMessage):
            if message.data.assignation in self.all_pod_assignments: 
                logger.info("Cancellation for task received. Canceling!")
                hosting_provising = self.all_pod_assignments[message.data.assignation]
                await self.provision_actor_queue_map[hosting_provising].put(message)
            else:
                logger.error("Received Cancellation for Assignation that is not running with us. Sending Done back..")
                unassign_done = UnassignDoneMessage(data={
                    "assignation": message.data.assignation
                }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
                await self.forward(unassign_done)

        else:
            logger.error(f"This Message shouldn't apear in the entertainer {message}")

                


    async def deactivateProvision(self, bounced_unprovide: BouncedUnprovideMessage):
        # Where should we do this?
        future = self.loop.create_future()
        self.unprovide_futures[bounced_unprovide.meta.reference] = future
        await self.forward(bounced_unprovide)
        reference = await future # is just a loopback??
        return reference

    async def activateProvision(self, bounced_provide: BouncedProvideMessage):
        
        #We are forwarding the Provision to Arkitekt and wait for its acknowledgement of the creation of this Pod
        future = self.loop.create_future()
        self.provide_futures[bounced_provide.meta.reference] = future
        await self.forward(bounced_provide)
        id = await future
        pod = await Pod.asyncs.get(id=id)
        return pod

    
    def actor_cancelled(self, actor: Actor, future: Future):
        logger.info("Actor is done! Cancellation or Finished??")
        if future.cancelled():
            logger.info("Future was cancelled everything is cools")


    async def get_actorclass_for_template(self, template_id):
        assert template_id in self.template_id_actorClass_map, "We have no Actor stored in our list"
        return self.template_id_actorClass_map[template_id]


    def registerActor(self, template_id: str, actorClass: Type[Actor]):
        assert template_id not in self.template_id_actorClass_map, "We cannot register two Actors for the same template on the same provider"
        self.template_id_actorClass_map[template_id] = actorClass


    async def unentertain(self, bounced_unprovide: BouncedUnprovideMessage):
        unprovide_handler = UnprovideHandler(bounced_unprovide, self.connector)

        try:
            provision_reference = bounced_unprovide.data.provision 
            assert provision_reference in self.provision_actor_run_map, "We dont entertain this provision"
            task = self.provision_actor_run_map[provision_reference]

            if not task.done():
                await unprovide_handler.pass_log("Cancelling", level=LogLevel.INFO)
                task.cancel()
                await unprovide_handler.pass_done()
            else:
                raise TaskAlreadyDoneError("Task was already done")

            
        except ProtocolException as e:
            console.print_exception()
            await unprovide_handler.pass_error(e)
        except Exception as e:
            console.print_exception()
            await unprovide_handler.pass_critical(e)
        
        
        # THIS Comes form the Arkitekt Platform

    
    async def entertain(self, bounced_provide: BouncedProvideMessage, actorClass: Type[Actor]):
        ''' Takes an instance of a pod, asks arnheim to activate it and accepts requests on it,
        cancel this task to unprovide your local implementatoin '''
        provision_reference = bounced_provide.meta.reference # We register Actors unter its provision
        actor = actorClass(self.connector)

        try:
            self.provision_actor_map[provision_reference] = actor
            self.provision_actor_queue_map[provision_reference] = actor.queue
            run_task = create_task(actor.run(bounced_provide))
            run_task.add_done_callback(partial(self.actor_cancelled, actor))
            self.provision_actor_run_map[provision_reference] = run_task

        except Exception as e:
            logger.error("We have a provision error")
            logger.error(e)
            # This error gets passed back to the provider
            raise e





from abc import ABC, abstractmethod
from asyncio.futures import Future
from contextvars import Context
from bergen.messages.postman.log import LogLevel
from bergen.messages import *
from bergen.messages.base import MessageModel
from bergen.postmans.utils import build_assign_message, build_reserve_message, build_unassign_messsage, build_unreserve_messsage
import uuid
from bergen.hookable.base import Hookable
from bergen.schema import Node, Pod, Template
from typing import Callable, Dict, List
from aiostream import stream
import asyncio
import logging
from bergen.console import console

logger = logging.getLogger(__name__)

ReferenceQueueMap = Dict[str, asyncio.Queue]
ReferenceFutureMap = Dict[str, Future]
ReferenceProgressFuncMap = Dict[str, Callable]

class NodeException(Exception):
    pass

class HostException(Exception):
    pass

class ProtocolException(Exception):
    pass

class BasePostman(Hookable):
    """ A Postman takes node requests and translates them to Bergen calls, basic implementations are GRAPHQL and PIKA"""
    
    def __init__(self, client, requires_configuration=True, loop=None,**kwargs) -> None:
        super().__init__(**kwargs)
        self.loop = loop or client.loop
        self.client = client

        # Assignments and their Cancellations
        self.assignment_stream_queues: ReferenceQueueMap = {}
        self.assignments: ReferenceFutureMap = {}
        self.assignment_progress_functions: ReferenceProgressFuncMap = {}

        self.unassignments: ReferenceFutureMap = {}
        self.unassignment_progress_functions: ReferenceProgressFuncMap = {}

        # Reservations and their Cancellations
        self.reservation_stream_queues: ReferenceQueueMap = {}
        self.reservations: ReferenceFutureMap = {}
        self.reservations_progress_functions: ReferenceProgressFuncMap = {}

        self.unreservations: ReferenceFutureMap = {}
        self.unreservations_progress_functions: ReferenceProgressFuncMap = {}


    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    @abstractmethod
    async def forward(self, message: MessageModel):
        return NotImplementedError("This is abstract")


    async def on_message(self, message: MessageModel):
        # First we check the streams
        reference = message.meta.reference

        if reference in self.assignment_stream_queues:
            return await self.assignment_stream_queues[reference].put(message)

        if reference in self.reservation_stream_queues:
            return await self.reservation_stream_queues[reference].put(message)

        else:
            console.log(f"Unknown message {message}") 
        

    async def stream_reserve_to_queue(self, queue, node_id: str = None, template_id: str = None , provision: str = None, params_dict: dict = {}, with_log= True, context=None):
        reserve_reference = str(uuid.uuid4())
        self.reservation_stream_queues[reserve_reference] = queue
        reserve = build_reserve_message(reserve_reference, node_id, template_id, provision, params_dict=params_dict, with_log=with_log, context=context)
        await self.forward(reserve)
        return reserve_reference

    async def send_unreserve(self, reservation: str = None, context: Context=None):
        unreserve_reference = str(uuid.uuid4())
        unreserve = build_unreserve_messsage(unreserve_reference, reservation, context=context)
        await self.forward(unreserve)
        return unreserve_reference

    async def delete_reservequeue(self, reference: str = None):
        del self.reservation_stream_queues[reference]

    async def stream_assign_to_queue(self, queue, reservation: str, shrinked_args, shrinked_kwargs = {}, with_log=True, persist=True, context = None):
        assign_reference = str(uuid.uuid4())
        self.assignment_stream_queues[assign_reference] = queue
        assign = build_assign_message(assign_reference, reservation, shrinked_args, kwargs=shrinked_kwargs, with_log=with_log, context=context, persist=persist)
        await self.forward(assign)
        return assign_reference

    async def send_unassign(self, assignation: str = None, context: Context= None):
        unassign_reference = str(uuid.uuid4())
        unreserve = build_unassign_messsage(unassign_reference, assignation, context=context)
        await self.forward(unreserve)
        return unassign_reference

    async def delete_assignqueue(self, reference: str = None):
        del self.assignment_stream_queues[reference]

    